# -*- coding: utf-8 -*-

"""Pick N Mix is a simple stacking tool for stacking Sci-Kit learn models of your picks.
It provided 2 classes: Layer and Stack. Layer is a parallel combination of models,
while Stack combine Layers to create a stacking model"""

from copy import deepcopy
import numpy as np

class Layer:
    def __init__(self, models, preprocessors=None, proba=False):
        """Initialize Layer, create a parallel combination of Sci-Kit learn models
        with or without preprocessors

        Parameters
        ==========
        preprocessors : a list of picks from sklearn.preprocessing,
                        if not None, the number of preprocessors and models must match.
                        If not using preprocessing for a model, None need to be put in place
        models : a list of picks from sklearn models
        proba : bool or a list of bool to show if predict_proba should be use instaed of predict,
                useful for classifiers not in the final Layer. If is a list,
                the length must match the number models.
        """
        if preprocessors is not None:
            assert len(preprocessors) == len(models), \
             "Number of preprocessors and models does not match, got {} processors but {} models.".format(len(preprocessors),len(models))

        if type(proba) != bool:
            assert len(proba) == len(models), \
             "Length of proba and number of models does not match, got {} processors but {} models.".format(len(proba),len(models))

        self.width = len(models) # number of models

        if preprocessors is None:
            self.preprocessors = [None] * self.width
        else:
            self.preprocessors = deepcopy(preprocessors)

        self.models = deepcopy(models)

        if type(proba) == bool:
            self.proba = [proba] * self.width
        else:
            self.proba = deepcopy(proba)

    def fit(self, X, y):
        """Fit each preprocessors and models in Layer with (X, y) and return
        predictions in an array of shape (n_samples, n_models) for the next Layer

        Parameters
        ==========
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Training data
        y : array_like, shape (n_samples, n_targets)
            Target values.

        Returns
        =======
        C : array, shape (n_samples, n_models)
            Returns predicted values for the next layer.
        """
        result = None
        for idx in range(self.width):
            if self.preprocessors[idx] is not None:
                X_new = self.preprocessors[idx].fit_transform(X)
            else:
                X_new = X

            self.models[idx].fit(X_new,y)

            if self.proba[idx]:
                temp_result = self.models[idx].predict_proba(X_new)
            else:
                temp_result = self.models[idx].predict(X_new)
                temp_result = np.expand_dims(temp_result, axis=1)

            if result is None:
                result = temp_result
            else:
                result = np.concatenate((result, temp_result), axis=1)
        return result

    def predict(self, X):
        """With put fiting any preprocessors and models in Layer, return predictions
        of X in an array of shape (n_samples, n_models) for the next Layer

        Parameters
        ==========
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples

        Returns
        =======
        C : array, shape (n_samples, n_models)
            Returns predicted values for the next layer.
        """
        result = None
        for idx in range(self.width):
            if self.preprocessors[idx] is not None:
                X_new = self.preprocessors[idx].transform(X)
            else:
                X_new = X

            if self.proba[idx]:
                temp_result = self.models[idx].predict_proba(X_new)
            else:
                temp_result = self.models[idx].predict(X_new)
                temp_result = np.expand_dims(temp_result, axis=1)

            if result is None:
                result = temp_result
            else:
                result = np.concatenate((result, temp_result), axis=1)
        return result


class Stack:
    def __init__(self, layers):
        """Initialize Stack, create a vertical stacking of Layers

        Parameters
        ==========
        layers : a list of Layers
        """
        self.depth = len(layers) # number of layers
        self.layers = deepcopy(layers)

    def fit(self, X, y):
        """Fit Layers with (X, y) and return the fitted Stack

        Parameters
        ==========
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Training data
        y : array_like, shape (n_samples, n_targets)
            Target values.

        Returns
        =======
        self : obejct, the fitted Stack itself
        """
        X_new = X
        for idx in range(self.depth):
            X_new = self.layers[idx].fit(X_new, y)
        return self # follow convention of Sci-Kit learn and return self

    def predict(self, X):
        """With given X, predict the result with the Stack

        Parameters
        ==========
        X : array-like or sparse matrix, shape (n_samples, n_features)
            Samples.

        Returns
        =======
        C : array, shape (n_samples,)
            Returns predicted values from the Stack.
        """
        X_new = X
        for idx in range(self.depth):
            X_new = self.layers[idx].predict(X_new)
        # flatten result if only a number for each X
        if X_new.shape[1] == 1:
            X_new = X_new.flatten()
        return X_new # this is the final result



import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LogisticRegression
from picknmix.picknmix import Layer, Stack

layer_width2_reg = Layer([LinearRegression(), LinearRegression()],
                         preprocessors = [None, MinMaxScaler()])
layer_width2_clf = Layer([LogisticRegression(solver='liblinear'),
                          LogisticRegression(solver='liblinear')],
                        proba=[True,False])
layer_width1_reg = Layer([LinearRegression()])
layer_width1_clf = Layer([LogisticRegression(solver='liblinear')])

class TestStack(object):

    def test_fit_predict_1_layer_reg(self):
        model = Stack([layer_width1_reg])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        model.fit(X, y)
        result = model.predict(np.array([[3, 5],[3, 5]]))
        assert result.shape == (2,)
        assert np.allclose(result, np.array([16, 16]))

    def test_fit_predict_1_layer_clf(self):
        model = Stack([layer_width1_clf])
        X = np.array([[1, 1], [1, 1], [0, 0], [0, 0]])
        y = np.array([1, 1, 0, 0])
        model.fit(X, y)
        result = model.predict(np.array([[1, 1]]))
        assert result.shape == (1,)
        assert np.allclose(result, np.array([1]))

    def test_fit_predict_2_layers_reg(self):
        model = Stack([layer_width2_reg, layer_width1_reg])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        model.fit(X, y)
        result = model.predict(np.array([[3, 5],[3, 5]]))
        assert result.shape == (2,)
        assert np.allclose(result, np.array([16, 16]))

    def test_fit_predict_2_layers_clf(self):
        model = Stack([layer_width2_clf, layer_width1_clf])
        X = np.array([[1, 1], [1, 1], [0, 0], [0, 0]])
        y = np.array([1, 1, 0, 0])
        model.fit(X, y)
        result = model.predict(np.array([[1, 1]]))
        assert result.shape == (1,)
        assert np.allclose(result, np.array([1]))



import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LogisticRegression
from picknmix.picknmix import Layer

class TestLayer(object):
    def test_different_numbers_of_preprocessor_and_models(self):
        with pytest.raises(Exception):
            assert Layer([LinearRegression(), LinearRegression()],
                         [MinMaxScaler()])

    def test_fit_single_model_without_preprocess(self):
        layer_model = Layer([LinearRegression()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        # X and y are linearly related, predictions will be almost perfect
        result = layer_model.fit(X, y)
        assert result.shape == (4,1)
        assert np.allclose(result.flatten(), y)

    def test_fir_single_model_with_preprocess(self):
        layer_model = Layer([LinearRegression()],
                            [MinMaxScaler()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        # X and y are linearly related, predictions will be almost perfect
        result = layer_model.fit(X, y)
        assert result.shape == (4,1)
        assert np.allclose(result.flatten(), y)

    def test_fit_single_model_with_2_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='liblinear')],
                            proba=True)
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 0])
        result = layer_model.fit(X, y)
        assert result.shape == (4,2)

    def test_fit_single_model_with_multi_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='lbfgs',
                                                multi_class='multinomial')],
                            proba=True)
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 2])

        result = layer_model.fit(X, y)
        assert result.shape == (4,3)

    def test_fit_multiple_models(self):
        layer_model = Layer([LinearRegression(), LinearRegression()],
                            [None, MinMaxScaler()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        # X and y are linearly related, predictions will be almost perfect
        result = layer_model.fit(X, y)
        assert result.shape == (4,2)
        assert np.allclose(result[:,0], y)
        assert np.allclose(result[:,1], y)

    def test_fit_multiple_model_with_2_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='liblinear'),
                             LogisticRegression(solver='liblinear')],
                            proba=[True,False])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 0])
        result = layer_model.fit(X, y)
        assert result.shape == (4,3)

    def test_predict_single_model_without_preprocess(self):
        layer_model = Layer([LinearRegression()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5],[3, 5]]))
        assert result.shape == (2,1)
        assert np.allclose(result, np.array([[16],[16]]))

    def test_predict_single_model_with_preprocess(self):
        layer_model = Layer([LinearRegression()],
                            [MinMaxScaler()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5]]))
        assert result.shape == (1,1)
        assert np.allclose(result, np.array([[16]]))

    def test_predict_single_model_with_2_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='liblinear')],
                            proba=True)
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 0])

        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5]]))
        assert result.shape == (1,2)

    def test_predict_single_model_with_multi_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='lbfgs',
                                                multi_class='multinomial')],
                            proba=True)
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 2])
        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5]]))
        assert result.shape == (1,3)

    def test_predict_multiple_model(self):
        layer_model = Layer([LinearRegression(), LinearRegression()],
                            [None, MinMaxScaler()])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.dot(X, np.array([1, 2])) + 3
        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5]]))
        assert result.shape == (1,2)
        assert np.allclose(result, np.array([[16, 16]]))

    def test_predict_multiple_model_with_2_class_proba(self):
        layer_model = Layer([LogisticRegression(solver='liblinear'),
                             LogisticRegression(solver='liblinear')],
                            proba=[True,False])
        X = np.array([[1, 1], [1, 2], [2, 2], [2, 3]])
        y = np.array([1, 1, 0, 0])
        layer_model.fit(X, y)
        result = layer_model.predict(np.array([[3, 5], [2, 5]]))
        assert result.shape == (2,3)
